class Requests {
    constructor() {
        this._items = new Map();
    }

    /**
     * Creates new request andd stores it in the list.
     *
     * @param {String|Number} requestId
     * @param {Function} fn
     * @param {Number} timeout
     * @returns {Promise}
     */
    create(requestId, fn, timeout) {
        this._rejectExistingRequest(requestId);
        return this._createNewRequest(requestId, fn, timeout);
    }

    resolve(requestId, data) {
        if (requestId && this._items.has(requestId)) {
            this._items.get(requestId).resolve(data);
        }
    }

    rejectAll(error) {
        this._items.forEach(request => request.isPending ? request.reject(error) : null);
    }

    _rejectExistingRequest(requestId) {
        const existingRequest = this._items.get(requestId);
        if (existingRequest && existingRequest.isPending) {
            existingRequest.reject(new Error(`WebSocket request is replaced, id: ${requestId}`));
        }
    }

    _createNewRequest(requestId, fn, timeout) {
        const request = new ControlledPromise();
        this._items.set(requestId, request);
        request.timeout(timeout, `WebSocket request was rejected by timeout (${timeout} ms). RequestId: ${requestId}`);
        // request._createPromise()
        // return promiseFinally(request.call(fn), () => this._deleteRequest(requestId, request));

        request._resolve = request.call(fn);
        request._createPromise();

        return request.promise;

        /*
                        return new Promise((resolve, reject) => {
                            console.log("_createNewRequest Promise fulfilled");
                            request.call(fn);
                            // this._deleteRequest(requestId, request);
                        });
        */

    }

    _deleteRequest(requestId, request) {
        // this check is important when request was replaced
        if (this._items.get(requestId) === request) {
            this._items.delete(requestId);
        }
    }
}

// end Requests

function flatOptions(options, defaults) {
    const result = Object.assign({}, defaults);
    if (options && typeof options === 'object') {
        Object.keys(options).forEach(key => validateOption(key, defaults) && copyOption(key, options, result));
    }
    return result;
}

function copyOption(key, from, to) {
    if (from[key] !== undefined) {
        to[key] = from[key];
    }
}

function validateOption(key, defaults) {
    if (defaults && !Object.hasOwnProperty.call(defaults, key)) {
        throw new Error(`Unknown option: ${key}`);
    }
    return true;
}

// end flat-options

let defaultOptions = {
        /**
         * See {@link Options.createWebSocket}
         *
         * @param {String} url
         * @returns {WebSocket}
         */
        createWebSocket: url => new WebSocket(url),

    /**
     * See {@link Options.packMessage}
     *
     * @param {*} data
     * @returns {String|ArrayBuffer|Blob}
     */
    packMessage: null,

    /**
     * See {@link Options.unpackMessage}
     *
     * @param {String|ArrayBuffer|Blob} message
     * @returns {*}
     */
    unpackMessage: null,

    /**
     * See {@link Options.attachRequestId}
     *
     * @param {*} data
     * @param {String|Number} requestId
     * @returns {*}
     */
    attachRequestId: null,

    /**
     * See {@link Options.extractRequestId}
     *
     * @param {*} data
     * @returns {String|Number|undefined}
     */
    extractRequestId: null,

    /**
     * See {@link Options.timeout}
     */
    timeout: 0,

    /**
     * See {@link Options.connectionTimeout}
     */
    connectionTimeout: 0,
};

// end defaultOptions

/**
 * Controlled Promise.
 */

class ControlledPromise {
    /**
     * Creates controlled promise. In contrast to original Promise, it does not immediately call any function.
     * Instead it has [.call()](#ControlledPromise+call) method for that and `resolve / reject` methods for
     * resolving promise.
     */
    constructor() {
        this._resolve = null;
        this._reject = null;
        this._isPending = false;
        this._isFulfilled = false;
        this._isRejected = false;
        this._value = undefined;
        this._promise = null;
        this._timer = null;
        this._timeout = 0;
        this._timeoutReason = 'Promise rejected by timeout';
    }

    /**
     * Returns promise itself.
     *
     * @returns {Promise}
     */
    get promise() {
        return this._promise;
    }

    /**
     * Returns value with that promise was fulfilled (resolved or rejected).
     *
     * @returns {*}
     */
    get value() {
        return this._value;
    }

    /**
     * Returns true if promise is pending.
     *
     * @returns {Boolean}
     */
    get isPending() {
        return this._isPending;
    }

    /**
     * Returns true if promise is fulfilled.
     *
     * @returns {Boolean}
     */
    get isFulfilled() {
        return this._isFulfilled;
    }

    /**
     * Returns true if promise rejected.
     *
     * @returns {Boolean}
     */
    get isRejected() {
        return this._isRejected;
    }

    /**
     * Returns true if promise fulfilled or rejected.
     *
     * @returns {Boolean}
     */
    get isSettled() {
        return this._isFulfilled || this._isRejected;
    }

    /**
     * Returns true if promise already called via `.call()` method.
     *
     * @returns {Boolean}
     */
    get isCalled() {
        return this.isPending || this.isSettled;
    }

    /**
     * This method executes `fn` and returns promise. While promise is pending all subsequent calls of `.call(fn)`
     * will return the same promise. To fulfill that promise you can use `.resolve() / .reject()` methods.
     *
     * @param {Function} fn
     * @returns {Promise}
     */
    call(fn) {
        if (!this._isPending) {
            this.reset();
            this._createPromise();
            this._callFn(fn);
            this._createTimer();
        }
        return this._promise;
    }

    /**
     * Resolves pending promise with specified `value`.
     *
     * @param {*} [value]
     */
    resolve(value) {
        if (this._isPending) {
            this._resolve(value);
        }
    }

    /**
     * Rejects pending promise with specified `value`.
     *
     * @param {*} [value]
     */
    reject(value) {
        if (this._isPending) {
            this._reject(value);
        }
    }

    /**
     * Resets to initial state.
     */
    reset() {
        if (this._isPending) {
            this.reject(new Error('Promise rejected by reset'));
        }
        this._promise = null;
        this._isPending = false;
        this._isFulfilled = false;
        this._isRejected = false;
        this._value = undefined;
        this._clearTimer();
    }

    /**
     * Sets timeout to reject promise automatically.
     *
     * @param {Number} ms delay in ms after that promise will be rejected automatically
     * @param {String|Error|Function} [reason] rejection value. If it is string or error - promise will be rejected with
     * that error. If it is function - this function will be called after delay where you can manually resolve or reject
     * promise via `.resolve() / .reject()` methods.
     */
    timeout(ms, reason) {
        this._timeout = ms;
        if (reason !== undefined) {
            this._timeoutReason = reason;
        }
    }

    _createPromise() {
        const internalPromise = new Promise((resolve, reject) => {
            this._isPending = true;
        this._resolve = resolve;
        this._reject = reject;
    });
        this._promise = internalPromise
            .then(value => this._handleFulfill(value), error => this._handleReject(error));
    }

    _handleFulfill(value) {
        this._settle(value);
        this._isFulfilled = true;
        return this._value;
    }

    _handleReject(value) {
        this._settle(value);
        this._isRejected = true;
        return Promise.reject(this._value);
    }

    _handleTimeout() {
        if (typeof this._timeoutReason === 'function') {
            this._timeoutReason();
        } else {
            const error = typeof this._timeoutReason === 'string'
                ? new Error(this._timeoutReason)
                : this._timeoutReason;
            this.reject(error);
        }
    }

    _createTimer() {
        if (this._timeout) {
            this._timer = setTimeout(() => this._handleTimeout(), this._timeout);
        }
    }

    _clearTimer() {
        if (this._timer) {
            clearTimeout(this._timer);
            this._timer = null;
        }
    }

    _settle(value) {
        this._isPending = false;
        this._value = value;
        this._clearTimer();
    }

    _callFn(fn) {
        if (typeof fn === 'function') {
            let result;
            try {
                result = fn();
            } catch (e) {
                this.reject(e);
            }
            this._tryAttachToPromise(result);
        }
    }

    _tryAttachToPromise(p) {
        const isPromise = p && typeof p.then === 'function';
        if (isPromise) {
            p.then(value => this.resolve(value), e => this.reject(e));
        }
    }
}


// end ControlledPromise

// channel.js
const innerEvents = [
    'onListenerAdded',
    'onListenerRemoved',
    'onFirstListenerAdded',
    'onLastListenerRemoved'
];

/**
 * Channel of particular events. Allows attach/detach listeners and dispatch event data.
 *
 * @param {String} [name]
 * @param {Boolean} [noInnerEvents]
 *
 * @example
 * import Channel from 'chnl';
 *
 * // create channel
 * const onMyEvent = new Channel();
 * // listen
 * onMyEvent.addListener(data => console.log(data));
 * // dispatch data
 * onMyEvent.dispatch(data);
 */
class Channel {
    constructor(name, noInnerEvents) {
        this._listeners = [];
        this._mute = false;
        this._accumulate = false;
        this._accumulatedEvents = [];
        this._name = name || '';
        this._noInnerEvents = Boolean(noInnerEvents);
        if (!noInnerEvents) {
            innerEvents.forEach(eventName => this[eventName] = new Channel(eventName, true));
        }
    }

    /**
     * Add listener for event
     * @param {Function} callback
     * @param {Object} [context]
     */
    addListener(callback, context) {
        this._pushListener(callback, context, false);
    }

    /**
     * Add once listener for event
     * @param {Function} callback
     * @param {Object} [context]
     */
    addOnceListener(callback, context) {
        this._pushListener(callback, context, true);
    }

    /**
     * Remove listener from event
     * @param {Function} callback
     * @param {Object} [context]
     */
    removeListener(callback, context) {
        this._ensureFunction(callback);
        const index = this._indexOfListener(callback, context);
        if (index >= 0) {
            this._spliceListener(index);
        }
    }

    /**
     * Remove all listeners from channel.
     */
    removeAllListeners() {
        while (this.hasListeners()) {
            this._spliceListener(0);
        }
    }

    /**
     * Is listener exist
     * @param {Function} callback
     * @param {Object} [context]
     * @returns {Boolean}
     */
    hasListener(callback, context) {
        this._ensureFunction(callback);
        return this._indexOfListener(callback, context) >= 0;
    }

    /**
     * Are there any listeners
     * @returns {Boolean}
     */
    hasListeners() {
        return this._listeners.length > 0;
    }

    /**
     * Call all listeners with specified params
     */
    dispatch(...args) {
    this._invokeListeners({args, async: false});
}

/**
 * Call all listeners with specified params asynchronously
 */
dispatchAsync(...args) {
    this._invokeListeners({args, async: true});
}

/**
 * Mute channel
 * @param {Object} [options]
 * @param {Boolean} [options.accumulate] accumulate events and call listeners after .unmute()
 */
mute(options = {}) {
    this._mute = true;
    if (options.accumulate) {
        this._accumulate = true;
    } else {
        this._accumulate = false;
        this._accumulatedEvents = [];
    }
}

/**
 * Unmute channel
 */
unmute() {
    this._mute = false;
    if (this._accumulate) {
        this._dispatchAccumulated();
        this._accumulate = false;
    }
}

/**
 * @param {Object} options
 * @param {Array} options.args
 * @param {Boolean} [options.async]
 * @private
 */
_invokeListeners(options = {args: [], async: false}) {
    if (!this._mute) {
        // ToDo: block adding/removing listeners to channel (throw an error) during dispatch operation
        const listnersToInvoke = this._listeners.slice();
        listnersToInvoke.forEach(listener => {
            this._invokeListener(listener, options);
        if (listener.once) {
            this.removeListener(listener.callback, listener.context)
        }
    });
    } else if (this._accumulate) {
        this._accumulatedEvents.push(options);
    }
}

/**
 * @param {Object} listener
 * @param {Object} options
 * @param {Array} options.args
 * @param {Boolean} options.async
 * @private
 */
_invokeListener(listener, options) {
    if (options.async) {
        setTimeout(() => listener.callback.apply(listener.context, options.args), 0);
    } else {
        listener.callback.apply(listener.context, options.args);
    }
}

/**
 * Ensure function
 * @param {Function} callback
 */
_ensureFunction(callback) {
    if (typeof callback !== 'function') {
        throw new Error('Channel ' + this._name + ': listener is not a function');
    }
}

/**
 * Dispatch inner events when listener is added
 * @private
 */
_dispatchInnerAddEvents() {
    if (!this._noInnerEvents) {
        this.onListenerAdded.dispatch.apply(this.onListenerAdded, arguments);
        if (this._listeners.length === 1) {
            this.onFirstListenerAdded.dispatch.apply(this.onFirstListenerAdded, arguments);
        }
    }
}

/**
 * Dispatch inner events when listener is removed
 * @private
 */
_dispatchInnerRemoveEvents() {
    if (!this._noInnerEvents) {
        this.onListenerRemoved.dispatch.apply(this.onListenerRemoved, arguments);
        if (this._listeners.length === 0) {
            this.onLastListenerRemoved.dispatch.apply(this.onLastListenerRemoved, arguments);
        }
    }
}

/**
 * Find listener index
 * @param {Function} callback
 * @param {Object} [context]
 * @private
 */
_indexOfListener(callback, context) {
    for (let i = 0; i < this._listeners.length; i++) {
        const listener = this._listeners[i];
        const equalCallbacks = listener.callback === callback;
        const emptyContexts = context === undefined && listener.context === undefined;
        const equalContexts = context === listener.context;
        if (equalCallbacks && (emptyContexts || equalContexts)) {
            return i;
        }
    }
}

/**
 * Dispatch accumulated events
 * @private
 */
_dispatchAccumulated() {
    this._accumulatedEvents.forEach(options => this._invokeListeners(options));
    this._accumulatedEvents = [];
}

/**
 * Pushes listener
 * @param {Function} callback
 * @param {Object} context
 * @param {Boolean} once
 */
_pushListener(callback, context, once) {
    this._ensureFunction(callback);
    this._listeners.push({callback, context, once});
    this._dispatchInnerAddEvents.apply(this, arguments);
}

/**
 * Splice listener under index
 * @param {Number} index
 */
_spliceListener(index) {
    const listener = this._listeners[index];
    this._listeners.splice(index, 1);
    const args = [listener.callback];
    if (listener.context) {
        args.push(listener.context);
    }
    this._dispatchInnerRemoveEvents.apply(this, args);
}
}
// end channel.js

// EventEmitter.js

/**
 * Event emitter similar to Node.js [EventEmitter](https://nodejs.org/api/events.html#events_class_eventemitter).
 * The main difference from single channel is that each method takes additional `event` argument.
 *
 * @example
 * import Channel from 'chnl';
 *
 * // create emitter
 * const emitter = new Channel.EventEmitter();
 * // listen 'myEvent'
 * emitter.on('myEvent', data => console.log(data));
 * // emit 'myEvent'
 * emitter.emit('myEvent', 'hello world!');
 */
class EventEmitter {
    constructor() {
        this._channels = new Map();
    }

    /**
     * Adds listener to specific event
     *
     * @param {String} event
     * @param {Function} callback
     * @param {Object} [context]
     */
    addListener(event, callback, context) {
        this._getChannel(event).addListener(callback, context);
    }

    /**
     * Adds listener to specific event (alias to addListener)
     *
     * @param {String} event
     * @param {Function} callback
     * @param {Object} [context]
     */
    on(event, callback, context) {
        this.addListener(event, callback, context);
    }

    /**
     * Adds once listener to specific event
     *
     * @param {String} event
     * @param {Function} callback
     * @param {Object} [context]
     */
    addOnceListener(event, callback, context) {
        this._getChannel(event).addOnceListener(callback, context);
    }

    /**
     * Adds once listener to specific event (alias to addOnceListener)
     *
     * @param {String} event
     * @param {Function} callback
     * @param {Object} [context]
     */
    once(event, callback, context) {
        this.addOnceListener(event, callback, context);
    }

    /**
     * Removes listener from specific event
     *
     * @param {String} event
     * @param {Function} callback
     * @param {Object} [context]
     */
    removeListener(event, callback, context) {
        this._getChannel(event).removeListener(callback, context);
    }

    /**
     * Removes listener from specific event (alias to removeListener)
     *
     * @param {String} event
     * @param {Function} callback
     * @param {Object} [context]
     */
    off(event, callback, context) {
        this.removeListener(event, callback, context);
    }

    /**
     * Is listener exist for specific event
     *
     * @param {String} event
     * @param {Function} callback
     * @param {Object} [context]
     * @returns {Boolean}
     */
    hasListener(event, callback, context) {
        return this._getChannel(event).hasListener(callback, context);
    }

    /**
     * Is listener exist for specific event (alias to hasListener)
     *
     * @param {String} event
     * @param {Function} callback
     * @param {Object} [context]
     * @returns {Boolean}
     */
    has(event, callback, context) {
        return this.hasListener(event, callback, context);
    }

    /**
     * Are there any listeners for specific event
     *
     * @returns {Boolean}
     */
    hasListeners(event) {
        return this._getChannel(event).hasListeners();
    }

    /**
     * Call all listeners for specific event
     *
     * @param {String} event
     * @param {*} args
     */
    dispatch(event, ...args) {
    this._getChannel(event).dispatch(...args);
}

/**
 * Call all listeners for specific event
 *
 * @param {String} event
 * @param {*} args
 */
emit(event, ...args) {
    this.dispatch(event, ...args);
}

/**
 * Returns channel by event name
 *
 * @param {String} event
 * @returns {Channel}
 * @private
 */
_getChannel(event) {
    if (!this._channels.has(event)) {
        this._channels.set(event, new Channel(event));
    }
    return this._channels.get(event);
}
}

// end EventEmitter.js


// flat-options.js
function flatOptions(options, defaults) {
    const result = Object.assign({}, defaults);
    if (options && typeof options === 'object') {
        Object.keys(options).forEach(key => validateOption(key, defaults) && copyOption(key, options, result));
    }
    return result;
}

function copyOption(key, from, to) {
    if (from[key] !== undefined) {
        to[key] = from[key];
    }
}

function validateOption(key, defaults) {
    if (defaults && !Object.hasOwnProperty.call(defaults, key)) {
        throw new Error(`Unknown option: ${key}`);
    }
    return true;
}
// end flat-options.js

// subscription-item.js

class SubscriptionItem {
    /**
     * Constructor
     *
     * @param {Object} params
     * @param {Object} params.channel
     * @param {String} [params.event]
     * @param {Function} params.listener
     */
    constructor(params) {
        this._params = params;
        this._isOn = false;
        this._assertParams();
    }

    /**
     * Turn on listener of channel
     */
    on() {
        if (!this._isOn) {
            const {channel} = this._params;
            const method = channel.addListener || channel.addEventListener || channel.on;
            this._applyMethod(method);
            this._isOn = true;
        }
    }

    /**
     * Turn off listener of channel
     */
    off() {
        if (this._isOn) {
            const {channel} = this._params;
            const method = channel.removeListener || channel.removeEventListener || channel.off;
            this._applyMethod(method);
            this._isOn = false;
        }
    }

    _applyMethod(method) {
        const {channel, event, listener} = this._params;
        const args = event ? [event, listener] : [listener];
        method.apply(channel, args);
    }

    _assertParams() {
        const {channel, event, listener} = this._params;
        if (!channel || typeof channel !== 'object') {
            throw new Error('Channel should be object');
        }
        if (event && typeof event !== 'string') {
            throw new Error('Event should be string');
        }
        if (!listener || typeof listener !== 'function') {
            throw new Error('Listener should be function');
        }
    }
}


// end subscription-item.js

// subscription.js

/**
 * Utility class allowing dynamically attach/detach batch of listeners to event channels.
 *
 * @param {Array<{channel, event, listener}>} items
 *
 * @example
 * import Channel from 'chnl';
 * const subscription = new Channel.Subscription([
 *   {
 *     channel: chrome.tabs.onUpdated,
 *     listener: this._onTabUpdated.bind(this)
 *   }
 * ]);
 *
 * // attach listeners
 * subscription.on();
 * // detach listeners
 * subscription.off();
 */
class Subscription {
    constructor(items) {
        this._items = items.map(params => new SubscriptionItem(params));
    }

    /**
     * Turn on all listeners
     *
     * @returns {Subscription}
     */
    on() {
        this._items.forEach(item => item.on());
        return this;
    }

    /**
     * Turn off all listeners
     *
     * @returns {Subscription}
     */
    off() {
        this._items.forEach(item => item.off());
        return this;
    }
}

// end subscription.js

// react-subscription.js
class ReactSubscription extends Subscription {
    constructor(component, items) {
        super(items);
        this._overrideComponentCallback(component, 'componentDidMount', 'on');
        this._overrideComponentCallback(component, 'componentWillUnmount', 'off');
    }

    /**
     * @param {ReactComponent} component
     * @param {String} callbackName
     * @param {String} methodName
     * @private
     */
    _overrideComponentCallback(component, callbackName, methodName) {
        const baseCallback = component[callbackName];
        component[callbackName] = (...args) => {
            this[methodName]();
            if (typeof baseCallback === 'function') {
                return baseCallback.apply(component, args);
            }
        };
    }
}
// end react-subscription.js

// chnl index.js

const chnl = Channel;
chnl.EventEmitter = EventEmitter;
chnl.Subscription = Subscription;
chnl.ReactSubscription = ReactSubscription

// end Channel

const STATE = {
    CONNECTING: 0,
    OPEN: 1,
    CLOSING: 2,
    CLOSED: 3,
};

/**
 * @typicalname wsp
 */
class WebSocketAsPromised {
    /**
     * Constructor. Unlike original WebSocket it does not immediately open connection.
     * Please call `open()` method to connect.
     *
     * @param {String} url WebSocket URL
     * @param {Options} [options]
     */
    constructor(url, options) {
        this._url = url;
        this._options = flatOptions(options, defaultOptions);
        this._opening = new ControlledPromise();
        this._closing = new ControlledPromise();
        this._requests = new Requests();
        this._onMessage = new Channel();
        this._onPackedMessage = new Channel();
        this._onResponse = new Channel();
        this._onClose = new Channel();
        this._ws = null;
        this._wsSubscription = null;
    }

    /**
     * Returns original WebSocket instance created by `options.createWebSocket`.
     *
     * @returns {WebSocket}
     */
    get ws() {
        return this._ws;
    }

    /**
     * Is WebSocket connection in opening state.
     *
     * @returns {Boolean}
     */
    get isOpening() {
        return Boolean(this._ws && this._ws.readyState === STATE.CONNECTING);
    }

    /**
     * Is WebSocket connection opened.
     *
     * @returns {Boolean}
     */
    get isOpened() {
        return Boolean(this._ws && this._ws.readyState === STATE.OPEN);
    }

    /**
     * Is WebSocket connection in closing state.
     *
     * @returns {Boolean}
     */
    get isClosing() {
        return Boolean(this._ws && this._ws.readyState === STATE.CLOSING);
    }

    /**
     * Is WebSocket connection closed.
     *
     * @returns {Boolean}
     */
    get isClosed() {
        return Boolean(!this._ws || this._ws.readyState === STATE.CLOSED);
    }

    /**
     * Event channel triggered every time when message received from server.
     *
     * @see https://vitalets.github.io/chnl/#channel
     * @example
     * wsp.onMessage.addListener(message => console.log(message));
     *
     * @returns {Channel}
     */
    get onMessage() {
        return this._onMessage;
    }

    /**
     * Event channel triggered every time when received message is successfully unpacked.
     *
     * @see https://vitalets.github.io/chnl/#channel
     * @example
     * wsp.onPackedMessage.addListener(data => console.log(data));
     *
     * @returns {Channel}
     */
    get onPackedMessage() {
        return this._onPackedMessage;
    }

    /**
     * Event channel triggered every time when requestId is found in received message.
     *
     * @see https://vitalets.github.io/chnl/#channel
     * @example
     * wsp.onResponse.addListener(data => console.log(data));
     *
     * @returns {Channel}
     */
    get onResponse() {
        return this._onResponse;
    }

    /**
     * Event channel triggered when connection closed.
     * Listener accepts single argument `{code, reason}`.
     *
     * @see https://vitalets.github.io/chnl/#channel
     *
     * @returns {Channel}
     */
    get onClose() {
        return this._onClose;
    }

    /**
     * Opens WebSocket connection. If connection already opened, promise will be resolved with "open event".
     *
     * @returns {Promise<Event>}
     */
    open() {
        if (this.isClosing) {
            return Promise.reject(new Error(`Can't open WebSocket while closing.`));
        }
        if (this.isOpened) {
            return this._opening.promise;
        }
        return this._opening.call(() => {
            const timeout = this._options.connectionTimeout || this._options.timeout;
        this._opening.timeout(timeout, `Can't open WebSocket within allowed timeout: ${timeout} ms.`);
        this._opening.promise.catch(e => this._cleanup(e));
        this._createWS();
    });
    }

    /**
     * Performs request and waits for response.
     *
     * @param {*} data
     * @param {Object} [options]
     * @param {String|Number} [options.requestId=<auto-generated>]
     * @param {Number} [options.timeout=0]
     * @returns {Promise}
     */
    sendRequest(data, options = {}) {
        const requestId = options.requestId || `${Math.random()}`;
        const timeout = options.timeout !== undefined ? options.timeout : this._options.timeout;
        return this._requests.create(requestId, () => {
            this._assertRequestIdHandlers();
        const finalData = this._options.attachRequestId(data, requestId);
        this.sendPacked(finalData);
    }, timeout);
    }

    /**
     * Packs data with `options.packMessage` and sends to the server.
     *
     * @param {*} data
     */
    sendPacked(data) {
        this._assertPackingHandlers();
        const message = this._options.packMessage(data);
        this.send(message);
    }

    /**
     * Sends data without packing.
     *
     * @param {String|Blob|ArrayBuffer} data
     */
    send(data) {
        if (this.isOpened) {
            this._ws.send(data);
        } else {
            throw new Error(`Can't send data because WebSocket is not opened.`);
        }
    }

    /**
     * Closes WebSocket connection. If connection already closed, promise will be resolved with "close event".
     *
     * @returns {Promise<Event>}
     */
    close() {
        if (this.isClosed) {
            return Promise.resolve(this._closing.value);
        }
        return this._closing.call(() => {
            const {timeout} = this._options;
        this._closing.timeout(timeout, `Can't close WebSocket within allowed timeout: ${timeout} ms.`);
        this._ws.close();
    });
    }

    _createWS() {
        this._ws = this._options.createWebSocket(this._url);
        this._wsSubscription = new Channel.Subscription([
            {channel: this._ws, event: 'open', listener: e => this._handleOpen(e)},
        {channel: this._ws, event: 'message', listener: e => this._handleMessage(e)},
        {channel: this._ws, event: 'error', listener: e => this._handleError(e)},
        {channel: this._ws, event: 'close', listener: e => this._handleClose(e)},
    ]).on();
    }

    _handleOpen(event) {
        this._opening.resolve(event);
    }

    _handleMessage(event) {
        const message = event.data;
        this._onMessage.dispatchAsync(message);
        this._handleUnpackedMessage(message);
    }

    _handleUnpackedMessage(message) {
        if (this._options.unpackMessage) {
            const data = this._options.unpackMessage(message);
            if (data !== undefined) {
                this._onPackedMessage.dispatchAsync(data);
                this._handleResponse(data);
            }
        }
    }

    _handleResponse(data) {
        if (this._options.extractRequestId) {
            const requestId = this._options.extractRequestId(data);
            if (requestId) {
                this._onResponse.dispatchAsync(data, requestId);
                this._requests.resolve(requestId, data);
            }
        }
    }

    _handleError() {
        // currently no specific handling of this event
    }

    _handleClose(event) {
        this._onClose.dispatchAsync(event);
        this._closing.resolve(event);
        const error = new Error(`WebSocket closed with reason: ${event.reason} (${event.code}).`);
        if (this._opening.isPending) {
            this._opening.reject(error);
        }
        this._cleanup(error);
    }

    _cleanupWS() {
        if (this._wsSubscription) {
            this._wsSubscription.off();
            this._wsSubscription = null;
        }
        this._ws = null;
    }

    _cleanup(error) {
        this._cleanupWS();
        this._requests.rejectAll(error);
    }

    _assertPackingHandlers() {
        if (!this._options.packMessage || !this._options.unpackMessage) {
            throw new Error(`Please define 'options.packMessage / options.unpackMessage' for sending packed messages.`);
        }
    }

    _assertRequestIdHandlers() {
        if (!this._options.attachRequestId || !this._options.extractRequestId) {
            throw new Error(`Please define 'options.attachRequestId / options.extractRequestId' for sending requests.`);
        }
    }
}
